// (C) 2008 Witold Filipczyk
#include <QApplication>
#include <QAction>
#include <QBitmap>
#include <QFileDialog>
#include <QLayout>
#include <QMainWindow>
#include <QMenu>
#include <QMenuBar>
#include <QMessageBox>
#include <QMouseEvent>
#include <QPushButton>
#include <QWidget>
#include <QObject>
#include <QCDEStyle>
#include <zlib.h>

const unsigned int char_width = 12;
const unsigned int char_height = 22;
const unsigned int char_size = char_height * ((char_width + 7) / 8);

class Editor : public QMainWindow
{
Q_OBJECT
	QWidget *w;
	QPushButton *siatka[char_height][char_width];
	QGridLayout *grid;
	QMenu *file, *help;
	QAction *openAction, *exitAction, *aboutAction, *aboutQtAction;
	unsigned char d[char_size];
	QString dir;
	public:
	Editor();
	void setChar(unsigned int width, unsigned int height, unsigned char data[]);
	unsigned char *getData(unsigned int width, unsigned int height);
	protected:
	virtual void closeEvent(QCloseEvent *);
	private slots:
	void open();
	void close();
	void about();
	void aboutQt();
};

Editor *edi;

class MojButton : public QPushButton
{
	unsigned int numer;
public:
	MojButton(unsigned int n, QWidget *w) : QPushButton(w), numer(n) {};
	unsigned int getNumer() { return numer; }
};

class Font : public QMainWindow
{
Q_OBJECT
	QWidget *w;
	MojButton **siatka;
	QGridLayout *grid;
	QMenu *file;
	QMenu *help;
	QAction *saveAction, *helpAction, *closeAction;
	QString dir;
	void changePixmap(unsigned int n);
	class PSF
	{
		public:
		unsigned int type;
		unsigned int unicodeLength;
		unsigned int restLength;
		unsigned char *data;
		unsigned char *unicodeData;
		unsigned char *rest;
		unsigned char header[4];
		unsigned char header2[28];
		unsigned int version, headersize, flags, length, charsize, height, width;
		PSF(const char *name);
		virtual ~PSF();
		void setData(unsigned int n, unsigned char d[16]);
		unsigned char *getData(unsigned int n);
	};
	PSF f;
public:
	Font(const char *name);
	virtual ~Font();
protected:
	bool eventFilter( QObject *obj, QEvent *ev );
private slots:
	void save();
	void info();
	void close();
};

Editor::Editor() : QMainWindow(), dir("/lib/kbd/consolefonts")
{
	w = new QWidget(this);
	file = menuBar()->addMenu("&File");
	help = menuBar()->addMenu("&Help");
	
	openAction = new QAction("&Open", this);
	connect(openAction, SIGNAL(triggered()), this, SLOT(open()));
	exitAction = new QAction("&Exit", this);
	connect(exitAction, SIGNAL(triggered()), this, SLOT(close()));
	file->addAction(openAction);
	file->addAction(exitAction);
	aboutAction = new QAction("&About", this);
	connect(aboutAction, SIGNAL(triggered()), this, SLOT(about()));
	aboutQtAction = new QAction("About &Qt", this);
	connect(aboutQtAction, SIGNAL(triggered()), this, SLOT(aboutQt()));
	help->addAction(aboutAction);
	help->addAction(aboutQtAction);

	grid = new QGridLayout(w);
	grid->setHorizontalSpacing(0);
	grid->setVerticalSpacing(0);
	for (unsigned int i = 0; i < char_height; i++)
		for (unsigned int j = 0; j < char_width; j++) {
			siatka[i][j] = new QPushButton(w);
			siatka[i][j]->setCheckable(true);
			siatka[i][j]->setMaximumSize(25, 25);
			grid->addWidget(siatka[i][j], i, j);
		}
	setCentralWidget(w);
	show();
}

void
Editor::open()
{
	QString name = QFileDialog::getOpenFileName(0, "Open PSF font", dir);
	if (name.isNull()) return;
	new Font(name.toAscii());
}

void
Editor::close()
{
	QApplication::exit();
}

void
Editor::closeEvent(QCloseEvent *)
{
	close();
}

void
Editor::about()
{	
	QMessageBox::about(this, "About AFE", "PSF font editor\n(C) 2008 Witold Filipczyk");
}

void
Editor::aboutQt()
{
	QMessageBox::aboutQt(this);
}

void
Editor::setChar(unsigned int width, unsigned int height, unsigned char data[])
{
	unsigned int charsize = (width + 7) / 8;
	height = height > char_height ? char_height : height;
	for (unsigned int i = 0; i < height; i++)
		for (unsigned int j = 0; j < charsize; j++) {
			unsigned int ind = i * charsize + j;
			for (int bit = 7; bit >= 0; bit--) {
				unsigned int indeks = 7 - bit + j * 8;
				if (indeks >= char_width) continue;
				if (data[ind] & (1 << bit))
					siatka[i][indeks]->setChecked(true);
				else
					siatka[i][indeks]->setChecked(false);
			}
		}
}

unsigned char *
Editor::getData(unsigned int width, unsigned int height)
{
	for (unsigned int i = 0; i < char_size; i++)
		d[i] = 0;

	height = height > char_height ? char_height : height;
	unsigned int size = (width + 7) / 8;
	for (unsigned int i = 0; i < height; i++)
		for (unsigned int j = 0; j < size; j++) {
			unsigned int ind = i * size + j;
			for (int bit = 7; bit >= 0; bit--) {
				unsigned int indeks = 7 - bit + j * 8;
				if (indeks >= char_width) continue;
				if (siatka[i][indeks]->isChecked())
					d[ind] |= (1 << bit);
			}
		}
	return d;
}

unsigned int
calc(unsigned char data[4])
{
	return data[0] + (data[1] << 8) + (data[2] << 16) + (data[3] << 24);
}

Font::PSF::PSF(const char *name):type(0), data(NULL), unicodeData(NULL), rest(NULL), length(0)
{
	gzFile gz = gzopen(name, "rb");
	int l = 0;

	if (!gz)
		return;
	gzread(gz, header, 4);
	if (header[0] == 0x36 && header[1] == 0x04) {	
		if (header[2] & 1)
			length = 512;
		else
			length = 256;
		if (header[2] & 2)
			l = 1;
		height = charsize = header[3];
		width = 8;
		data = new unsigned char[length * charsize];
		gzread(gz, data, length * charsize);
		type = 1;
	} else if (header[0] == 0x72 && header[1] == 0xb5 && header[2] == 0x4a && header[3] == 0x86) {
		gzread(gz, header2, 28);
		version = calc(header2);
		headersize = calc(header2 + 4);
		flags = calc(header2 + 8);
		length = calc(header2 + 12);
		charsize = calc(header2 + 16);
		height = calc(header2 + 20);
		width = calc(header2 + 24);
		l = flags & 1;
		restLength = headersize - 32;
		if (restLength > 0) {
			rest = new unsigned char[restLength];
			gzread(gz, rest, restLength);
		}
		data = new unsigned char[length * charsize];
		gzread(gz, data, length * charsize);
		type = 2;
	}
	unicodeLength = 0;
	while (l) {
		unicodeData = (unsigned char *)realloc(unicodeData, unicodeLength + 65536);
		l = gzread(gz, unicodeData + unicodeLength, 65536);
		if (l <= 0) break;
		unicodeLength += l;
	}
	gzclose(gz);
}

void
Font::PSF::setData(unsigned int n, unsigned char d[])
{
	unsigned int ind = n * charsize;

	for (unsigned int i = 0; i < charsize; i++)
		data[ind + i] = d[i];
}

unsigned char *
Font::PSF::getData(unsigned int n)
{
	return data + (n * charsize);
}

Font::PSF::~PSF()
{
//fprintf(stderr, "Before koniec PSF\n");
	if (type) {
		delete[] data;
		if (unicodeData) free(unicodeData);
		delete[] rest;
	}
//fprintf(stderr, "After koniec\n");
}

void
Font::changePixmap(unsigned int n)
{
	QSize size(f.width, f.height);
	QBitmap q = QBitmap::fromData(size, f.getData(n), QImage::Format_Mono);
	siatka[n]->setIcon(q);
	siatka[n]->setMaximumSize(f.width + 4, f.height + 4);
} 

Font::Font(const char *name) : QMainWindow(), dir(getenv("HOME")), f(name)
{
	if (!f.type) {
		siatka = 0;
		w = 0;
		QMessageBox::critical(this, "Error", QString(name) + " is not a PSF file!"); 
		return;
	}
	file = menuBar()->addMenu("&File");
	help = menuBar()->addMenu("&Help");
 	saveAction = new QAction("&Save as", this);
	connect(saveAction, SIGNAL(triggered()), this, SLOT(save()));
	file->addAction(saveAction);
	helpAction = new QAction("&Help", this);
	connect(helpAction, SIGNAL(triggered()), this, SLOT(info()));
	help->addAction(helpAction);
	closeAction = new QAction("&Close", this);
	file->addAction(closeAction);
	connect(closeAction, SIGNAL(triggered()), this, SLOT(close()));
	w = new QWidget(this);
	setWindowTitle(name);
	grid = new QGridLayout(w);
	grid->setHorizontalSpacing(2);
	grid->setVerticalSpacing(2);
	unsigned int i;
	siatka = new MojButton *[f.length];
	for (i = 0; i < f.length; i++) {
		unsigned int ii = i / 32;
		unsigned int j = i % 32;
		siatka[i] = new MojButton(i, w);
		changePixmap(i);
		siatka[i]->installEventFilter(this);
		grid->addWidget(siatka[i], ii, j);
	}
	setCentralWidget(w);
	show();
}

Font::~Font()
{
//	for (unsigned int i = 0; i < f.length; i++) siatka[i] = 0;
	delete siatka;
}

bool
Font::eventFilter( QObject *obj, QEvent *ev)
{
	if (ev->type() == QEvent::MouseButtonPress) {
		MojButton *but = reinterpret_cast<MojButton *>(obj);
		int number = but->getNumer();
		QMouseEvent *me = (QMouseEvent *)ev;
		Qt::MouseButton b = me->button();
		unsigned char *d;

		switch (b) {
		case Qt::LeftButton:
			edi->setChar(f.width, f.height, f.getData(number));
			break;
		case Qt::MidButton:
			f.setData(number, edi->getData(f.width, f.height));
			changePixmap(number);
			break;
		case Qt::RightButton:
			d = edi->getData(f.width, f.height);
			edi->setChar(f.width, f.height, f.getData(number));
			f.setData(number, d);
			changePixmap(number);
		default:
			break;
		}
		return true;
	} else {
		return false;
	}
	
}

void
Font::save()
{
	QString filename = QFileDialog::getSaveFileName(0, "Save", dir);
	if (filename.isNull()) return;
	const char *name = filename.toAscii();
	if (filename.endsWith(".gz")) {
		gzFile gz = gzopen(name, "wb");
		if (!gz) {
			QMessageBox::critical(this, "Error", "Cannot write " + filename + " !"); 
			return;
		}
		if (f.type == 1) {
			gzwrite(gz, f.header, 4);
			gzwrite(gz, f.data, (f.charsize * f.length));
			gzwrite(gz, f.unicodeData, f.unicodeLength);
		} else if (f.type == 2) {
			gzwrite(gz, f.header, 4);
			gzwrite(gz, f.header2, 28);
			gzwrite(gz, f.rest, f.restLength);
			gzwrite(gz, f.data, (f.charsize * f.length));
			gzwrite(gz, f.unicodeData, f.unicodeLength);
		}
		gzclose(gz);
	} else {
		FILE *fi = fopen(name, "wb");
		if (!fi) {
			QMessageBox::critical(this, "Error", "Cannot write " + filename + " !"); 
			return;
		}
		if (f.type == 1) {
			fwrite(f.header, 1, 4, fi);
			fwrite(f.data, 1, (f.charsize * f.length), fi);
			fwrite(f.unicodeData, 1, f.unicodeLength, fi);
		} else if (f.type == 2) {
			fwrite(f.header, 1, 4, fi);
			fwrite(f.header2, 1, 28, fi);
			fwrite(f.rest, 1, f.restLength, fi);
			fwrite(f.data, 1, (f.charsize * f.length), fi);
			fwrite(f.unicodeData, 1, f.unicodeLength, fi);
		}
		fclose(fi);
	}
}

void
Font::info()
{
	QMessageBox::information(this, "Help", "Left mouse button: edit\n"
	"Middle mouse button: paste\nRight mouse button: swap");	
}

void
Font::close()
{
	QMainWindow::close();
}

int main(int argc, char *argv[])
{
	QApplication app(argc, argv);
	app.setStyleSheet("QPushButton:checked { background-color: black }");
	edi = new Editor;
	for (int i = 1; i < argc; i++)
		new Font(argv[i]);
	return app.exec();
}

#include "afe.moc"
